
#define _GNU_SOURCE

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <dlfcn.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <netdb.h>
#include <errno.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <setjmp.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/uio.h>

#include "tesla-internal.h"
#include "tesla.h"

// Path to the master application.  LIBDIR must be defined on the command line
#define TESLA_MASTER_PATH LIBDIR "/tesla/teslamaster"

// Used by TS_CALL_LIBC and TS_APPLY_LIBC
void *ts_libc;

// The following functions are redefined by TESLA
int accept(int fd, struct sockaddr *addr, socklen_t *addrlen);
int bind(int fd, const struct sockaddr *addr, socklen_t addrlen);
int close(int fd);
int connect(int fd, const struct sockaddr *addr, socklen_t addrlen);
int listen(int fd, int backlog);
int socket(int domain, int type, int protocol);

int getsockopt(int s, int level, int optname, void *optval, socklen_t *optlen);
int setsockopt(int s, int level, int optname, const void *optval,
	       socklen_t optlen);

// These are overloaded by TESLA only for SOCK_DGRAM sockets
int read(int fd, void *buf, size_t len);
int recv(int fd, void *buf, size_t len, int flags);
int recvfrom(int fd, void *buf, size_t count, int flags,
	      struct sockaddr *from, socklen_t *fromlen);
int write(int fd, const void *buf, size_t len);
int send(int fd, const void *buf, size_t len, int flags);
int sendto(int fd, const void *buf, size_t count, int flags,
	   const struct sockaddr *to, socklen_t tolen);

// These should be overloaded by TESLA in the future
int recvmsg(int fd, struct msghdr *msg, int flags);
int sendmsg(int fd, const struct msghdr *msg, int flags);


/* The Tesla sessions array. */
ts_fd_info ts_fds[FD_SETSIZE];

static int master_fd = -1;
static unsigned int next_msg_id = 1;

int get_fd_from_master();


int read_fully(int fd, void *buf, size_t count)
{
    int orig_count = count;
    while (count) {
        int bytes = read(fd, buf, count);
        if (bytes <= 0)
            return 0;
        assert(bytes <= count);
        buf += bytes;
        count -= bytes;
    }
    return orig_count;
}

int write_fully(int fd, const void *buf, size_t count)
{
    int orig_count = count;
    while (count) {
        int written = write(fd, buf, count);
        if (written <= 0)
            return 0;
        assert(written <= count);
        buf += written;
        count -= written;
    }
    return orig_count;
}



unsigned int send_master(master_msg_t *msg) {
    msg->id = next_msg_id++;
    msg->pid = getpid();
    msg->len = sizeof *msg - sizeof msg->len;

    ts_debug_2("Sending message to master: id=%d, pid=%d, len=%d, type=%s",
	       msg->id, msg->pid, msg->len, MSG_TYPE(msg->type));

    if (write_fully(master_fd, msg, sizeof *msg) != sizeof *msg)
	ts_fatal("Unable to send message to master: %s", strerror(errno));

    return msg->id;
}

unsigned int send_master_header(master_msg_t *msg, int len) {
    msg->len = MASTER_MSG_HDRLEN + len - sizeof msg->len;
    msg->id = next_msg_id++;
    msg->pid = getpid();

    ts_debug_2("Sending partial message to master: id=%d, pid=%d, len=%d, type=%s",
	       msg->id, msg->pid, msg->len, MSG_TYPE(msg->type));

    if (write_fully(master_fd, msg, MASTER_MSG_HDRLEN) != MASTER_MSG_HDRLEN)
    {
	ts_fatal("Unable to send message to master: %s", strerror(errno));
    }

    return msg->id;
}

unsigned int recv_master(master_msg_t *msg) {
    int bytes;

    ts_debug_2("Reading length from master");
    if ((bytes = read_fully(master_fd, &msg->len, sizeof msg->len)) != sizeof msg->len)
	ts_fatal("Expected length from master but got %d bytes", bytes);
    ts_debug_2("Reading data (%d bytes) from master", msg->len);
    if ((bytes = read_fully(master_fd, &msg->id, msg->len)) != msg->len)
	ts_fatal("Expected %d bytes from master but got %d bytes", msg->len, bytes);

    ts_debug_2("Received message: id=%d, length=%d, type=%s", msg->id, msg->len, MSG_TYPE(msg->type));
    if (msg->type == MSG_FH_FOLLOWS) {
	ts_debug_2(" - waiting for FD (FH id was %d)", msg->conn_id);
	msg->body.fh_pass.fh = get_fd_from_master();
	ts_debug_2(" - got FD");
    }

    return msg->id;
}

void cleanup(void)
{
    master_msg_t msg;

    ts_debug_2("In cleanup");
    msg.type = MSG_NAK;
    send_master(&msg);
}

static void init() __attribute__((constructor));
static void init()
{
    master_msg_t msg;
    int sv[2];
    int socket;
    unsigned int id;
    pid_t pid;

    ts_libc = 0;

    socket = socketpair(AF_UNIX, SOCK_STREAM, 0, sv);
    if (socket < 0)
	ts_fatal("Unable to create socketpair: %s", strerror(errno));

    if ((pid = fork()) != 0) {
	/* Master process.  Pass control on to teslamaster.pm. */
	char pidstr[32];
	char fdstr[32];

	close(sv[0]);

	sprintf(pidstr, "%d", pid);
	sprintf(fdstr, "%d", sv[1]);

	unsetenv("LD_PRELOAD");

	if (getenv("TS_GDB")) {

	  FILE *f;

	  f = fopen("args", "w");
	  fprintf(f, "run %s %s\n", pidstr, fdstr);
	  fclose(f);

	  execlp("gdb", "gdb", TESLA_MASTER_PATH, "-x", "args", 0);
	} else {
	    execlp(TESLA_MASTER_PATH, TESLA_MASTER_PATH, pidstr, fdstr, 0);
	}

	ts_fatal("Unable to exec teslamaster: %s", strerror(errno));
    }

    close(sv[1]);
    master_fd = sv[0];

    /* Child process. */
    msg.type = MSG_HELLO;
    id = send_master(&msg);

    if (recv_master(&msg) != id || msg.type != MSG_HELLO)
	ts_fatal("Did not receive HELLO from master, %d, %d", id, msg.type);

    atexit(cleanup);
}

#define	CONTROLLEN	(sizeof(struct cmsghdr) + sizeof(int))
int
get_fd_from_master() {
    /* From Perl File::FDpasser */
    static struct cmsghdr *cmptr = 0;

    int newfd = -1, nread, status;
    char *ptr, buf[4096];
    struct iovec iov[1];
    struct msghdr msg;
  
    status = -1;
    iov[0].iov_base = buf;
    iov[0].iov_len  = sizeof(buf);
    msg.msg_iov     = iov;
    msg.msg_iovlen  = 1;
    msg.msg_name    = NULL;
    msg.msg_namelen = 0;
    if (cmptr == NULL && (cmptr = malloc(CONTROLLEN)) == NULL) return(-1);
    msg.msg_control    = (caddr_t) cmptr;
    msg.msg_controllen = CONTROLLEN;
    
    if ( (nread = recvmsg(master_fd, &msg, 0)) < 0) return(-1);
    else if (nread == 0)
	ts_fatal("Connection closed by master while reading filehandle");

    for (ptr = buf; ptr < &buf[nread]; ) {
	if (*ptr++ == 0) {
	    if (ptr != &buf[nread-1]) { ts_fatal("message format error"); }
	    status = *ptr & 255;
	    if (status == 0) {
		if (msg.msg_controllen != CONTROLLEN) { ts_fatal("status = 0 but no fd"); }
		newfd = *(int *)CMSG_DATA(cmptr); /* new descriptor */
	    } else newfd = -status;
	    nread -= 2;
	}
    }
    if (nread > 0) if (write(STDERR_FILENO, buf, nread) != nread) { ts_fatal("???"); }
    if (status >= 0) return(newfd);       /* final data has arrived,  descriptor, or -status */

    return 0;
}


int socket(int domain, int type, int protocol)
{
    /* Ask the master to give us a socket. */
    master_msg_t msg;
    unsigned int id;

    ts_debug_3("socket(domain=%d, type=%d, protocol=%d)", domain, type, protocol);

    msg.type = MSG_SOCKET;
    msg.body.socket.domain = domain;
    msg.body.socket.type = type;
    msg.body.socket.protocol = protocol;
    id = send_master(&msg);

    if (recv_master(&msg) != id)
	ts_fatal("Mis-sequenced response");
    if (msg.type == MSG_NAK) {
	ts_debug_3(" - declined by master");
	TS_APPLY_LIBC(socket);
    } else if (msg.type == MSG_FH_FOLLOWS) {
	int fd = msg.body.fh_pass.fh;
	assert(fd >= 0 && fd < FD_SETSIZE);
	ts_fds[fd].wrapped = 1;
	ts_fds[fd].domain = domain;
	ts_fds[fd].type = type;
	ts_fds[fd].conn_id = msg.conn_id;
	ts_debug_3(" - got FD %d from master; ID %d", fd, msg.conn_id);
	return fd;
    } else {
	ts_fatal("Expected NAK or FH_FOLLOWS but got %d", msg.type);
    }
}

int connect(int fd, const struct sockaddr *addr, socklen_t addrlen)
{
    master_msg_t msg;
    unsigned int id;
    char buf;
    size_t bytes;
    
    if (!ts_fds[fd].wrapped)
	TS_APPLY_LIBC(connect);

    msg.conn_id = ts_fds[fd].conn_id;
    msg.type = MSG_CONNECT;
    assert(addrlen <= sizeof msg.body.address.addr);
    msg.body.address.addrlen = addrlen;
    memcpy(msg.body.address.addr, addr, addrlen);
    id = send_master(&msg);

    /* Wait for connect by reading on socket. XXX: Use async magic */
    ts_debug_3("Connect - waiting for response");

    if (recv_master(&msg) != id)
	ts_fatal("Mismatched message");

    if (msg.type != MSG_ACK) {
	ts_debug_3(" - connect itself failed!");
	errno = msg.merrno;
	return -1;
    }

    if (ts_fds[fd].type == SOCK_DGRAM)
	return 0;

    bytes = TS_CALL_LIBC(read, fd, &buf, 1);
    if (bytes != 1)
	ts_fatal("Didn't read one byte after connect");
	
    if (buf == 'A') {
	ts_debug_2("Connect failed");
	errno = ECONNREFUSED;
	return -1;
    } else if (buf == 'B') {
	ts_debug_2("Connect succeeded");
	return 0;
    } else {
	ts_fatal("Expected A or B for connect; got %u", (unsigned int)buf);
    }
}

int listen(int fd, int backlog)
{
    master_msg_t msg;

    if (!ts_fds[fd].wrapped)
	TS_APPLY_LIBC(listen);

    ts_debug_2("Listening!");

    msg.conn_id = ts_fds[fd].conn_id;
    msg.type = MSG_LISTEN;
    send_master(&msg);

    return 0;
}

int bind(int fd, const struct sockaddr *addr, socklen_t addrlen)
{
    master_msg_t msg;
    unsigned int id;

    if (!ts_fds[fd].wrapped)
	TS_APPLY_LIBC(bind);
    
    msg.conn_id = ts_fds[fd].conn_id;
    msg.type = MSG_BIND;
    assert(addrlen <= sizeof msg.body.address.addr);
    msg.body.address.addrlen = addrlen;
    memcpy(msg.body.address.addr, addr, addrlen);
    id = send_master(&msg);

    if (recv_master(&msg) != id)
	ts_fatal("Mismatched message");

    if (msg.type == MSG_ACK) {
	ts_debug_2("Bind succeeded");
	return 0;
    } else if (msg.type == MSG_NAK) {
	ts_debug_2("Bind failed");
	errno = msg.merrno;
	return -1;
    }

    ts_fatal("Expected ACK or NAK but got %d", msg.type);
}

int accept(int fd, struct sockaddr *addr, socklen_t *addrlen)
{
    master_msg_t msg;
    unsigned int id;
    unsigned char buf;
    size_t bytes;

    if (!ts_fds[fd].wrapped)
	TS_APPLY_LIBC(accept);

    ts_debug_2("In accept... reading from master");

    /* Read a byte from the FD. If it's blocking, it will wait for a byte (C) */
    bytes = TS_CALL_LIBC(read, fd, &buf, 1);

    ts_debug_2(" - got %u bytes", bytes);

    if (bytes == 1) {
	if (buf != 'C')
	    ts_fatal("Expected C but got %u", (unsigned int)buf);
	
	ts_debug_2("Got a 'C' for accept");

	msg.conn_id = ts_fds[fd].conn_id;
	msg.type = MSG_ACCEPT;
	id = send_master(&msg);    

	if (recv_master(&msg) != id)
	    ts_fatal("Mismatched message");

	if (msg.type == MSG_ACK) {
	    *addrlen = msg.body.address.addrlen;
	    memcpy(addr, msg.body.address.addr, msg.body.address.addrlen);

	    if (recv_master(&msg) != id)
		ts_fatal("Mismatched message");
	    if (msg.type != MSG_FH_FOLLOWS)
		ts_fatal("Expected filehandle");

	    ts_fds[msg.body.fh_pass.fh].wrapped = 1;
	    ts_fds[msg.body.fh_pass.fh].conn_id = msg.conn_id;

	    ts_debug_2("Returning FD %d", msg.body.fh_pass.fh);

	    return msg.body.fh_pass.fh;
	}

	if (msg.type != MSG_NAK)
	    ts_fatal("Mismatched message");

	return -1;
    }
    else
	return -1;
}


int __close(int fd)
{
  int ret;

  if (!ts_fds[fd].wrapped)
    TS_APPLY_LIBC(__close);

  ret = TS_CALL_LIBC(__close, fd);
  ts_fds[fd].wrapped = 0;
  return ret;
}



int close(int fd)
{
  int ret;

  if (!ts_fds[fd].wrapped)
    TS_APPLY_LIBC(close);

  ret = TS_CALL_LIBC(close, fd);
  ts_fds[fd].wrapped = 0;
  return ret;
}



/* We need to relay these to the master process */

int
setsockopt(int fd, int level, int optname, const void *optval,
	   socklen_t optlen)
{
    unsigned int id;
    master_msg_t msg;
    const ts_ioctl_s *io = 0;

    ts_debug_2("Beginning of setsockopt");

    if (level == SOL_TESLA_IOCTL) {
	ts_debug_2(" - it's an ioctl");

	io = optval;

	ts_debug_2("ioctl: target=%s, optname=%d, optlen=%d, retlen=%p",
		   io->target, io->optname, io->optlen, io->retlen);
		   
	if (io == 0 || optlen != sizeof(ts_ioctl_s)) {
	    ts_error("Invalid arguments to IOCTL in setsockopt");
	    errno = EINVAL;
	    return -1;
	}
	assert(strlen(io->target) < sizeof msg.body.sockopt.target);
	strcpy(msg.body.sockopt.target, io->target);
	optname = io->optname;
	optval = io->optval;
	optlen = io->optlen;
    }

    if (!ts_fds[fd].wrapped)
	TS_APPLY_LIBC(setsockopt);

    msg.conn_id = ts_fds[fd].conn_id;
    msg.type = MSG_SETSOCKOPT;
    msg.body.sockopt.level = level;
    msg.body.sockopt.optname = optname;
    msg.body.sockopt.optlen = optlen;

    ts_debug_1("In setsockopt: %d %d", level, optname);

    switch(level) {
      default:
	ts_debug_2("Sending header");
	id = send_master_header(&msg, optlen + sizeof msg.body.sockopt - sizeof msg.body.sockopt.optval);
	ts_debug_2("Sending option info");
	if (write_fully(master_fd, &msg.body.sockopt, sizeof msg.body.sockopt - sizeof msg.body.sockopt.optval) != (sizeof msg.body.sockopt - sizeof msg.body.sockopt.optval))
	    ts_fatal("unable to send setsockopt data");
	ts_debug_2("Sending option data");
	if (write_fully(master_fd, optval, optlen) != optlen)
	    ts_fatal("unable to send setsockopt data");
	ts_debug_2("Getting ack/nak");

	if (recv_master(&msg) != id)
	    ts_fatal("Mismatched message");

	errno = msg.merrno;
	if (msg.type != MSG_ACK)
	    return -1;

	if (io) {
	    if (msg.body.sockopt.optlen < *(io->retlen))
		*(io->retlen) = msg.body.sockopt.optlen;
	    memcpy(io->retval, msg.body.sockopt.optval, *(io->retlen));
	}

	return 0;
    }

    /* Not reached */
    return 0;
}

int
getsockopt(int fd, int level, int optname, void *optval, socklen_t *optlen)
{
    unsigned int id;
    master_msg_t msg;

    if (!ts_fds[fd].wrapped)
	TS_APPLY_LIBC(getsockopt);
  
    msg.conn_id = ts_fds[fd].conn_id;
    msg.type = MSG_GETSOCKOPT;
    msg.body.sockopt.level = level;
    msg.body.sockopt.optname = optname;
    msg.body.sockopt.optlen = *optlen;

    ts_debug_1("In getsockopt");
    switch(level) {
      default:
	id = send_master(&msg);
	if (recv_master(&msg) != id)
	    ts_fatal("Mismatched message");

	if (msg.type == MSG_ACK) {
	    memcpy(optval, msg.body.sockopt.optval, *optlen = msg.body.sockopt.optlen);
	    return 0;
	}

	*optlen = 0;
	return -1;
    }

    /* Not reached */
    return 0;
}


int read(int fd, void *buf, size_t len) {
    if (!ts_fds[fd].wrapped || ts_fds[fd].type != SOCK_DGRAM)
	TS_APPLY_LIBC(read);
    return recvfrom(fd, buf, len, 0, 0, 0);
}


int recv(int fd, void *buf, size_t len, int flags) {
    if (!ts_fds[fd].wrapped || ts_fds[fd].type != SOCK_DGRAM)
	TS_APPLY_LIBC(recv);
    return recvfrom(fd, buf, len, flags, 0, 0);
}


int
recvfrom(int fd, void *buf, size_t count, int flags,
	 struct sockaddr *from, socklen_t *fromlen)
{
    static char addr[ADDRESS_SIZE + sizeof(socklen_t)];
    static struct msghdr msg;
    static struct iovec iov[4];
    int ret;

    if (!ts_fds[fd].wrapped || ts_fds[fd].type != SOCK_DGRAM)
	TS_APPLY_LIBC(recvfrom);

    if (flags)
	ts_fatal("TESLA doesn't support recvfrom/sendto flags");

    if (from && fromlen) {
	int cnt = 0;
	int orig_fromlen = *fromlen;

	msg.msg_name = 0;
	msg.msg_namelen = 0;
	msg.msg_iov = iov;
	iov[cnt].iov_base = fromlen;
	iov[cnt].iov_len = sizeof *fromlen;
	++cnt;
	iov[cnt].iov_base = from;
	iov[cnt].iov_len = *fromlen < ADDRESS_SIZE ? *fromlen : ADDRESS_SIZE;
	++cnt;
	if (*fromlen < ADDRESS_SIZE) {
	    iov[cnt].iov_base = addr;
	    iov[cnt].iov_len = ADDRESS_SIZE - *fromlen;
	    ++cnt;
	}
	iov[cnt].iov_base = buf;
	iov[cnt].iov_len = count;
	++cnt;
	msg.msg_iovlen = cnt;
	msg.msg_control = 0;
	msg.msg_controllen = 0;
	msg.msg_flags = 0;

	ret = TS_CALL_LIBC(recvmsg, fd, &msg, 0);
	if (*fromlen > orig_fromlen)
	    *fromlen = orig_fromlen;
    } else {
	msg.msg_name = 0;
	msg.msg_namelen = 0;
	msg.msg_iov = iov;
	iov[0].iov_base = addr;
	iov[0].iov_len = ADDRESS_SIZE + sizeof(socklen_t);
	iov[1].iov_base = buf;
	iov[1].iov_len = count;
	msg.msg_iovlen = 2;
	msg.msg_control = 0;
	msg.msg_controllen = 0;
	msg.msg_flags = 0;

	ret = TS_CALL_LIBC(recvmsg, fd, &msg, 0);
    }

    if (ret < ADDRESS_SIZE + sizeof(socklen_t)) {
	ts_error("recvmsg failed; message too short");
	return 0;
    }

    return ret - (ADDRESS_SIZE + sizeof(socklen_t));
}


int write(int fd, const void *buf, size_t len) {
    if (!ts_fds[fd].wrapped || ts_fds[fd].type != SOCK_DGRAM)
	TS_APPLY_LIBC(write);
    return sendto(fd, buf, len, 0, 0, 0);
}


int send(int fd, const void *buf, size_t len, int flags) {
    if (!ts_fds[fd].wrapped || ts_fds[fd].type != SOCK_DGRAM)
	TS_APPLY_LIBC(send);
    return sendto(fd, buf, len, flags, 0, 0);
}


int
sendto(int fd, const void *buf, size_t count, int flags,
       const struct sockaddr *to, socklen_t tolen)
{
    static char garbage[ADDRESS_SIZE];
    static struct msghdr msg;
    static struct iovec iov[4];

    int cnt;
    int ret;

    if (fd == 0) {
	ts_error("FD is 0!");
    }

    if (!ts_fds[fd].wrapped || ts_fds[fd].type != SOCK_DGRAM)
	TS_APPLY_LIBC(sendto);

    if (flags)
	ts_fatal("TESLA doesn't support recvfrom/sendto flags");
    
    // Essentially the same code in teslamaster::avail; consolidate?
    cnt = 0;
    msg.msg_name = 0;
    msg.msg_namelen = 0;
    msg.msg_iov = iov;
    iov[cnt].iov_base = &tolen;
    iov[cnt].iov_len = sizeof tolen;
    ++cnt;
    if (to && tolen) {
	ts_debug_2("have an address");
	iov[cnt].iov_base = (void*)to;
	iov[cnt].iov_len = tolen > MAX_OPTVAL_SIZE ? MAX_OPTVAL_SIZE : tolen;
	++cnt;
    }
    if (tolen < ADDRESS_SIZE) {
	ts_debug_3("address not long enough - padding");
	iov[cnt].iov_base = garbage;
	iov[cnt].iov_len = ADDRESS_SIZE - tolen;
	++cnt;
    }

    ts_debug_3("data length is %d", count);
    iov[cnt].iov_base = (void*)buf;
    iov[cnt].iov_len = count;
    ++cnt;

    ts_debug_3("iovlen is %d", msg.msg_iovlen);
    msg.msg_iovlen = cnt;
    msg.msg_control = 0;
    msg.msg_controllen = 0;
    msg.msg_flags = 0;

    ret = TS_CALL_LIBC(sendmsg, fd, &msg, 0);

    if (ret == 0)
	return 0;
    if (ret < 0) {
	ts_error("Sendmsg failed: %s", strerror(errno));
	return -1;
    }
    if (ret == sizeof(socklen_t) + ADDRESS_SIZE + count) {
	errno = 0;
	return count;
    }
    
    ts_error("Sendmsg - partial send");
    errno = EINVAL;
    return -1;
}


/* XXX: The recv/sendmsg commands are much trickier */

int
recvmsg(int fd, struct msghdr *msg, int flags)
{
  if (!ts_fds[fd].wrapped)
    TS_APPLY_LIBC(recvmsg);

  ts_fatal("TESLA doesn't yet support sendmsg/recvmsg");
  return -1;
}

int
sendmsg(int fd, const struct msghdr *msg, int flags)
{
  if (!ts_fds[fd].wrapped)
    TS_APPLY_LIBC(getsockopt);

  ts_fatal("TESLA doesn't yet support sendmsg/recvmsg");
  return -1;
}

static int
getxname(int fd, struct sockaddr *addr, socklen_t *addrlen, int which)
{
    unsigned int id;
    master_msg_t msg;
  
    msg.conn_id = ts_fds[fd].conn_id;
    msg.type = which;

    id = send_master(&msg);
    if (recv_master(&msg) != id)
	ts_fatal("Mismatched message");

    if (msg.type == MSG_ACK) {
	*addrlen = *addrlen < msg.body.address.addrlen ? *addrlen : msg.body.address.addrlen;
	memcpy(addr, msg.body.address.addr, *addrlen);
	return 0;
    }

    return -1;
}

int
getpeername(int fd, struct sockaddr *addr, socklen_t *addrlen)
{
    if (!ts_fds[fd].wrapped)
	TS_APPLY_LIBC(getpeername);

    return getxname(fd, addr, addrlen, MSG_GETPEERNAME);
}

int
getsockname(int fd, struct sockaddr *addr, socklen_t *addrlen)
{
    if (!ts_fds[fd].wrapped)
	TS_APPLY_LIBC(getsockname);

    return getxname(fd, addr, addrlen, MSG_GETSOCKNAME);
}
